@@ -1,5 +1,5 @@
 # rawdog: RSS aggregator without delusions of grandeur.
-# Copyright 2003, 2004 Adam Sampson <azz@us-lot.org>
+# Copyright 2003, 2004, 2005 Adam Sampson <ats@offog.org>
 #
 # rawdog is free software; you can redistribute and/or modify it
 # under the terms of that license as published by the Free Software
@@ -12,15 +12,16 @@
 # General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with rawdog; see the file COPYING. If not, write to the
-# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
-# MA 02111-1307 USA, or see http://www.gnu.org/.
+# along with rawdog; see the file COPYING. If not, write to the Free
+# Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
+# MA 02110-1301, USA, or see http://www.gnu.org/.
 
-VERSION = "2.1"
+VERSION = "2.6"
 STATE_VERSION = 2
-import feedparser, feedfinder
+import feedparser, feedfinder, plugins
 from persister import Persistable, Persister
-import os, time, sha, getopt, sys, re, urlparse, cgi, socket, urllib2, calendar
+import os, time, sha, getopt, sys, re, cgi, socket, urllib2, calendar
+import string, threading
 from StringIO import StringIO
 
 def set_socket_timeout(n):
@@ -73,7 +74,7 @@ def sanitise_html(html, baseurl, inline, config):
 	p.feed(html)
 	html = p.output()
 
-	if not inline:
+	if not inline and config["blocklevelhtml"]:
 		# If we're after some block-level HTML and the HTML doesn't
 		# start with a block-level element, then insert a <p> tag
 		# before it. This still fails when the HTML contains text, then
@@ -92,7 +93,9 @@ def sanitise_html(html, baseurl, inline, config):
 		html = output[output.find("<body>") + 6
 		              : output.rfind("</body>")].strip()
 
-	return html
+	box = plugins.Box(html)
+	plugins.call_hook("clean_html", config, box, baseurl, inline)
+	return box.value
 
 def select_detail(details):
 	"""Pick the preferred type of detail from a list of details. (If the
@@ -138,39 +141,66 @@ def detail_to_html(details, inline, config, force_preformatted = False):
 
 	return sanitise_html(html, detail["base"], inline, config)
 
-def url_to_html(url):
-	"""Convert a URL string to HTML."""
-	return cgi.escape(url)
+def author_to_html(entry, feedurl, config):
+	"""Convert feedparser author information to HTML."""
+	author_detail = entry.get("author_detail")
 
-template_re = re.compile(r'__(.*?)__')
+	if author_detail is not None and author_detail.has_key("name"):
+		name = author_detail["name"]
+	else:
+		name = entry.get("author")
+
+	url = None
+	if author_detail is not None:
+		if author_detail.has_key("url"):
+			url = author_detail["url"]
+		elif author_detail.has_key("email"):
+			url = "mailto:" + author_detail["email"]
+
+	if url is None:
+		html = name
+	else:
+		html = "<a href=\"" + cgi.escape(url) + "\">" + cgi.escape(name) + "</a>"
+
+	# We shouldn't need a base URL here anyway.
+	return sanitise_html(html, feedurl, True, config)
+
+def string_to_html(s, config):
+	"""Convert a string to HTML."""
+	return sanitise_html(cgi.escape(s), "", True, config)
+
+template_re = re.compile(r'(__.*?__)')
 def fill_template(template, bits):
 	"""Expand a template, replacing __x__ with bits["x"], and only
 	including sections bracketed by __if_x__ .. [__else__ ..]
 	__endif__ if bits["x"] is not "". If not bits.has_key("x"),
 	__x__ expands to ""."""
+	result = plugins.Box()
+	plugins.call_hook("fill_template", template, bits, result)
+	if result.value is not None:
+		return result.value
+
 	f = StringIO()
-	l = template_re.split(template)
-	i = 0
 	if_stack = []
-	while 1:
+	def write(s):
 		if not False in if_stack:
-			f.write(l[i])
-		i += 1
-		if i == len(l):
-			break
-		key = l[i]
-		if key.startswith("if_"):
-			k = key[3:]
-			if_stack.append(bits.has_key(k) and bits[k] != "")
-		elif key == "endif":
-			if if_stack != []:
-				if_stack.pop()
-		elif key == "else":
-			if if_stack != []:
-				if_stack.append(not if_stack.pop())
-		elif bits.has_key(key):
-			f.write(bits[key])
-		i += 1
+			f.write(s)
+	for part in template_re.split(template):
+		if part.startswith("__") and part.endswith("__"):
+			key = part[2:-2]
+			if key.startswith("if_"):
+				k = key[3:]
+				if_stack.append(bits.has_key(k) and bits[k] != "")
+			elif key == "endif":
+				if if_stack != []:
+					if_stack.pop()
+			elif key == "else":
+				if if_stack != []:
+					if_stack.append(not if_stack.pop())
+			elif bits.has_key(key):
+				write(bits[key])
+		else:
+			write(part)
 	return f.getvalue()
 
 file_cache = {}
@@ -183,10 +213,41 @@ def load_file(name):
 		f.close()
 	return file_cache[name]
 
+def write_ascii(f, s, config):
+	"""Write the string s, which should only contain ASCII characters, to
+	file f; if it isn't encodable in ASCII, then print a warning message
+	and write UTF-8."""
+	try:
+		f.write(s)
+	except UnicodeEncodeError, e:
+		config.bug("Error encoding output as ASCII; UTF-8 has been written instead.\n", e)
+		f.write(s.encode("UTF-8"))
+
 def short_hash(s):
 	"""Return a human-manipulatable 'short hash' of a string."""
 	return sha.new(s).hexdigest()[-8:]
 
+def decode_structure(struct, encoding):
+	"""Walk through a structure returned by feedparser, decoding any
+	strings that haven't already been converted to Unicode."""
+	def is_dict(t):
+		return (t is dict) or (t is feedparser.FeedParserDict)
+	for (key, value) in struct.items():
+		if type(value) is str:
+			try:
+				struct[key] = value.decode(encoding)
+			except:
+				# If the encoding's invalid, at least preserve
+				# the byte stream.
+				struct[key] = value.decode("ISO-8859-1")
+		elif is_dict(type(value)):
+			decode_structure(value, encoding)
+		elif type(value) is list:
+			for item in value:
+				if is_dict(type(item)):
+					decode_structure(item, encoding)
+
+non_alphanumeric_re = re.compile(r'<[^>]*>|\&[^\;]*\;|[^a-z0-9]')
 class Feed:
 	"""An RSS feed."""
 
@@ -206,13 +267,9 @@ class Feed:
 			return 0
 		else:
 			return 1
-	
-	def update(self, rawdog, now, config):
-		"""Fetch articles from a feed and add them to the collection.
-		Returns 1 if any articles were read, 0 otherwise."""
 
-		articles = rawdog.articles
-		handlers = []
+	def fetch(self, rawdog, config):
+		"""Fetch the current set of articles from the feed."""
 
 		class DummyPasswordMgr:
 			def __init__(self, creds):
@@ -222,46 +279,56 @@ class Feed:
 			def find_user_password(self, realm, authuri):
 				return self.creds
 
+		handlers = []
+
 		if self.args.has_key("user") and self.args.has_key("password"):
 			mgr = DummyPasswordMgr((self.args["user"], self.args["password"]))
 			handlers.append(urllib2.HTTPBasicAuthHandler(mgr))
 
 		proxies = {}
-		for key in self.args.keys():
+		for key, arg in self.args.items():
 			if key.endswith("_proxy"):
-				proxies[key[:-6]] = self.args[key]
-		if len(proxies.keys()) != 0:
+				proxies[key[:-6]] = arg
+		if len(proxies) != 0:
 			handlers.append(urllib2.ProxyHandler(proxies))
 
 		if self.args.has_key("proxyuser") and self.args.has_key("proxypassword"):
 			mgr = DummyPasswordMgr((self.args["proxyuser"], self.args["proxypassword"]))
 			handlers.append(urllib2.ProxyBasicAuthHandler(mgr))
 
-		feedparser._FeedParserMixin.can_contain_relative_uris = []
-		feedparser._FeedParserMixin.can_contain_dangerous_markup = []
+		plugins.call_hook("add_urllib2_handlers", rawdog, config, self, handlers)
+
 		try:
-			p = feedparser.parse(self.url,
+			return feedparser.parse(self.url,
 				etag = self.etag,
 				modified = self.modified,
 				agent = "rawdog/" + VERSION,
 				handlers = handlers)
-			status = p.get("status")
 		except:
-			p = None
-			status = None
+			return None
 
+	def update(self, rawdog, now, config, p):
+		"""Add new articles from a feed to the collection.
+		Returns True if any articles were read, False otherwise."""
+
+		status = None
+		if p is not None:
+			status = p.get("status")
 		self.last_update = now
 
 		error = None
-		non_fatal = 0
+		non_fatal = False
 		old_url = self.url
 		if p is None:
-			error = "Error parsing feed."
-		elif status is None:
+			error = "Error fetching or parsing feed."
+		elif status is None and len(p["feed"]) == 0:
 			if config["ignoretimeouts"]:
-				return 0
+				return False
 			else:
 				error = "Timeout while reading feed."
+		elif status is None:
+			# Fetched by some protocol that doesn't have status.
+			pass
 		elif status == 301:
 			# Permanent redirect. The feed URL needs changing.
 
@@ -272,7 +339,7 @@ class Feed:
 				error += "The config file has been updated automatically."
 			else:
 				error += "You should update its entry in your config file."
-			non_fatal = 1
+			non_fatal = True
 		elif status in [403, 410]:
 			# The feed is disallowed or gone. The feed should be unsubscribed.
 			error = "The feed has gone.\n"
@@ -282,6 +349,8 @@ class Feed:
 			error = "The feed returned an error.\n"
 			error += "If this condition persists, you should remove it from your config file."
 
+		plugins.call_hook("feed_fetched", rawdog, config, self, p, error, non_fatal)
+
 		if error is not None:
 			print >>sys.stderr, "Feed:        " + old_url
 			if status is not None:
@@ -289,7 +358,9 @@ class Feed:
 			print >>sys.stderr, error
 			print >>sys.stderr
 			if not non_fatal:
-				return 0
+				return False
+
+		decode_structure(p, p.get("encoding") or "UTF-8")
 
 		self.etag = p.get("etag")
 		self.modified = p.get("modified")
@@ -298,35 +369,44 @@ class Feed:
 		# and feed will be empty. In this case we return 0 so that
 		# we know not to expire articles that came from this feed.
 		if len(p["entries"]) == 0:
-			return 0
+			return False
 
 		self.feed_info = p["feed"]
 		feed = self.url
+		articles = rawdog.articles
 
-		if config["currentonly"]:
-			for (hash, a) in articles.items():
-				if a.feed == feed:
-					del articles[hash]
-
+		seen = {}
 		sequence = 0
 		for entry_info in p["entries"]:
 			article = Article(feed, entry_info, now, sequence)
+			ignore = plugins.Box(False)
+			plugins.call_hook("article_seen", rawdog, config, article, ignore)
+			if ignore.value:
+				continue
+			seen[article.hash] = True
 			sequence += 1
 
 			if articles.has_key(article.hash):
 				articles[article.hash].update_from(article, now)
+				plugins.call_hook("article_updated", rawdog, config, article, now)
 			else:
 				articles[article.hash] = article
+				plugins.call_hook("article_added", rawdog, config, article, now)
 
-		return 1
+		if config["currentonly"]:
+			for (hash, a) in articles.items():
+				if a.feed == feed and not seen.has_key(hash):
+					del articles[hash]
+
+		return True
 
 	def get_html_name(self, config):
 		if self.feed_info.has_key("title_detail"):
 			r = detail_to_html(self.feed_info["title_detail"], True, config)
 		elif self.feed_info.has_key("link"):
-			r = url_to_html(self.feed_info["link"])
+			r = string_to_html(self.feed_info["link"], config)
 		else:
-			r = url_to_html(self.url)
+			r = string_to_html(self.url, config)
 		if r is None:
 			r = ""
 		return r
@@ -334,10 +414,17 @@ class Feed:
 	def get_html_link(self, config):
 		s = self.get_html_name(config)
 		if self.feed_info.has_key("link"):
-			return '<a href="' + url_to_html(self.feed_info["link"]) + '">' + s + '</a>'
+			return '<a href="' + string_to_html(self.feed_info["link"], config) + '">' + s + '</a>'
 		else:
 			return s
 
+	def get_id(self, config):
+		if self.args.has_key("id"):
+			return self.args["id"]
+		else:
+			r = self.get_html_name(config).lower()
+			return non_alphanumeric_re.sub('', r)
+
 class Article:
 	"""An article retrieved from an RSS feed."""
 
@@ -435,9 +522,9 @@ def parse_time(value, default = "m"):
 	the default argument to change this. Raises ValueError if the format
 	isn't recognised."""
 	units = { "s" : 1, "m" : 60, "h" : 3600, "d" : 86400, "w" : 604800 }
-	for unit in units.keys():
+	for unit, size in units.items():
 		if value.endswith(unit):
-			return int(value[:-len(unit)]) * units[unit]
+			return int(value[:-len(unit)]) * size
 	return int(value) * units[default]
 
 def parse_bool(value):
@@ -455,14 +542,38 @@ def parse_list(value):
 	"""Parse a list of keywords separated by whitespace."""
 	return value.strip().split(None)
 
+def parse_feed_args(argparams, arglines):
+	"""Parse a list of feed arguments. Raise ConfigError if the syntax is invalid."""
+	args = {}
+	for a in argparams:
+		as = a.split("=", 1)
+		if len(as) != 2:
+			raise ConfigError("Bad feed argument in config: " + a)
+		args[as[0]] = as[1]
+	for a in arglines:
+		as = a.split(None, 1)
+		if len(as) != 2:
+			raise ConfigError("Bad argument line in config: " + a)
+		args[as[0]] = as[1]
+	if "maxage" in args:
+		args["maxage"] = parse_time(args["maxage"])
+	return args
+
 class ConfigError(Exception): pass
 
 class Config:
 	"""The aggregator's configuration."""
 
 	def __init__(self):
+		self.files_loaded = []
+		self.loglock = threading.Lock()
+		self.reset()
+
+	def reset(self):
 		self.config = {
 			"feedslist" : [],
+			"feeddefaults" : {},
+			"defines" : {},
 			"outputfile" : "output.html",
 			"maxarticles" : 200,
 			"maxage" : 0,
@@ -479,53 +590,80 @@ class Config:
 			"ignoretimeouts" : 0,
 			"daysections" : 1,
 			"timesections" : 1,
+			"blocklevelhtml" : 1,
 			"tidyhtml" : 0,
 			"sortbyfeeddate" : 0,
 			"currentonly" : 0,
 			"hideduplicates" : "",
 			"newfeedperiod" : "3h",
 			"changeconfig": 0,
+			"numthreads": 0,
 			}
 
 	def __getitem__(self, key): return self.config[key]
 	def __setitem__(self, key, value): self.config[key] = value
 
-	def load(self, filename):
+	def reload(self):
+		self.log("Reloading config files")
+		self.reset()
+		for filename in self.files_loaded:
+			self.load(filename, False)
+
+	def load(self, filename, explicitly_loaded = True):
 		"""Load configuration from a config file."""
+		if explicitly_loaded:
+			self.files_loaded.append(filename)
+
+		lines = []
 		try:
 			f = open(filename, "r")
-			lines = f.readlines()
+			for line in f.xreadlines():
+				stripped = line.strip()
+				if stripped == "" or stripped[0] == "#":
+					continue
+				if line[0] in string.whitespace:
+					if lines == []:
+						raise ConfigError("First line in config cannot be an argument")
+					lines[-1][1].append(stripped)
+				else:
+					lines.append((stripped, []))
 			f.close()
 		except IOError:
 			raise ConfigError("Can't read config file: " + filename)
-		for line in lines:
-			line = line.strip()
+
+		for line, arglines in lines:
 			try:
-				self.load_line(line)
+				self.load_line(line, arglines)
 			except ValueError:
 				raise ConfigError("Bad value in config: " + line)
 
-	def load_line(self, line):
-		"""Process a configuration line."""
-
-		if line == "" or line[0] == "#":
-			return
+	def load_line(self, line, arglines):
+		"""Process a configuration directive."""
 
 		l = line.split(None, 1)
-		if len(l) != 2:
+		if len(l) == 1 and l[0] == "feeddefaults":
+			l.append("")
+		elif len(l) != 2:
 			raise ConfigError("Bad line in config: " + line)
 
+		handled_arglines = False
 		if l[0] == "feed":
 			l = l[1].split(None)
 			if len(l) < 2:
 				raise ConfigError("Bad line in config: " + line)
-			args = {}
-			for a in l[2:]:
-				as = a.split("=", 1)
-				if len(as) != 2:
-					raise ConfigError("Bad feed argument in config: " + a)
-				args[as[0]] = as[1]
-			self["feedslist"].append((l[1], parse_time(l[0]), args))
+			self["feedslist"].append((l[1], parse_time(l[0]), parse_feed_args(l[2:], arglines)))
+			handled_arglines = True
+		elif l[0] == "feeddefaults":
+			self["feeddefaults"] = parse_feed_args(l[1].split(None), arglines)
+			handled_arglines = True
+		elif l[0] == "define":
+			l = l[1].split(None, 1)
+			if len(l) != 2:
+				raise ConfigError("Bad line in config: " + line)
+			self["defines"][l[0]] = l[1]
+		elif l[0] == "plugindirs":
+			for dir in parse_list(l[1]):
+				plugins.load_plugins(dir, self)
 		elif l[0] == "outputfile":
 			self["outputfile"] = l[1]
 		elif l[0] == "maxarticles":
@@ -558,6 +696,8 @@ class Config:
 			self["daysections"] = parse_bool(l[1])
 		elif l[0] == "timesections":
 			self["timesections"] = parse_bool(l[1])
+		elif l[0] == "blocklevelhtml":
+			self["blocklevelhtml"] = parse_bool(l[1])
 		elif l[0] == "tidyhtml":
 			self["tidyhtml"] = parse_bool(l[1])
 		elif l[0] == "sortbyfeeddate":
@@ -570,15 +710,33 @@ class Config:
 			self["newfeedperiod"] = l[1]
 		elif l[0] == "changeconfig":
 			self["changeconfig"] = parse_bool(l[1])
+		elif l[0] == "numthreads":
+			self["numthreads"] = int(l[1])
 		elif l[0] == "include":
-			self.load(l[1])
+			self.load(l[1], False)
+		elif plugins.call_hook("config_option_arglines", self, l[0], l[1], arglines):
+			handled_arglines = True
+		elif plugins.call_hook("config_option", self, l[0], l[1]):
+			pass
 		else:
 			raise ConfigError("Unknown config command: " + l[0])
 
+		if arglines != [] and not handled_arglines:
+			raise ConfigError("Bad argument lines in config after: " + line)
+
 	def log(self, *args):
 		"""If running in verbose mode, print a status message."""
 		if self["verbose"]:
+			self.loglock.acquire()
 			print >>sys.stderr, "".join(map(str, args))
+			self.loglock.release()
+
+	def bug(self, *args):
+		"""Report detection of a bug in rawdog."""
+		print >>sys.stderr, "Internal error detected in rawdog:"
+		print >>sys.stderr, "".join(map(str, args))
+		print >>sys.stderr, "This could be caused by a bug in rawdog itself or in a plugin."
+		print >>sys.stderr, "Please send this error message and your config file to the rawdog author."
 
 def edit_file(filename, editfunc):
 	"""Edit a file in place: for each line in the input file, call
@@ -587,8 +745,7 @@ def edit_file(filename, editfunc):
 	newname = "%s.new-%d" % (filename, os.getpid())
 	oldfile = open(filename, "r")
 	newfile = open(newname, "w")
-	for line in oldfile.xreadlines():
-		editfunc(line, newfile)
+	editfunc(oldfile, newfile)
 	newfile.close()
 	oldfile.close()
 	os.rename(newname, filename)
@@ -596,14 +753,12 @@ def edit_file(filename, editfunc):
 class AddFeedEditor:
 	def __init__(self, feedline):
 		self.feedline = feedline
-		self.seen = False
-	def edit(self, line, outputfile):
-		if not self.seen:
-			ls = line.split(None, 1)
-			if len(ls) > 0 and ls[0] == "feed":
-				outputfile.write(self.feedline)
-				self.seen = True
-		outputfile.write(line)
+	def edit(self, inputfile, outputfile):
+		d = inputfile.read()
+		outputfile.write(d)
+		if not d.endswith("\n"):
+			outputfile.write("\n")
+		outputfile.write(self.feedline)
 
 def add_feed(filename, url, config):
 	"""Try to add a feed to the config file."""
@@ -620,11 +775,66 @@ class ChangeFeedEditor:
 	def __init__(self, oldurl, newurl):
 		self.oldurl = oldurl
 		self.newurl = newurl
-	def edit(self, line, outputfile):
-		ls = line.strip().split(None)
-		if len(ls) > 2 and ls[0] == "feed" and ls[2] == self.oldurl:
-			line = line.replace(self.oldurl, self.newurl, 1)
-		outputfile.write(line)
+	def edit(self, inputfile, outputfile):
+		for line in inputfile.xreadlines():
+			ls = line.strip().split(None)
+			if len(ls) > 2 and ls[0] == "feed" and ls[2] == self.oldurl:
+				line = line.replace(self.oldurl, self.newurl, 1)
+			outputfile.write(line)
+
+class FeedFetcher:
+	"""Class that will handle fetching a set of feeds in parallel."""
+
+	def __init__(self, rawdog, feedlist, config):
+		self.rawdog = rawdog
+		self.config = config
+		self.lock = threading.Lock()
+		self.jobs = {}
+		for feed in feedlist:
+			self.jobs[feed] = 1
+		self.results = {}
+
+	def worker(self, num):
+		rawdog = self.rawdog
+		config = self.config
+
+		config.log("Thread ", num, " starting")
+		while 1:
+			self.lock.acquire()
+			if self.jobs == {}:
+				job = None
+			else:
+				job = self.jobs.keys()[0]
+				del self.jobs[job]
+			self.lock.release()
+			if job is None:
+				break
+
+			config.log("Thread ", num, " fetching feed: ", job)
+			feed = rawdog.feeds[job]
+			plugins.call_hook("pre_update_feed", rawdog, config, feed)
+			self.results[job] = feed.fetch(rawdog, config)
+		config.log("Thread ", num, " done")
+
+	def run(self, numworkers):
+		self.config.log("Thread farm starting with ", len(self.jobs), " jobs")
+		workers = []
+		for i in range(numworkers):
+			self.lock.acquire()
+			isempty = (self.jobs == {})
+			self.lock.release()
+			if isempty:
+				# No jobs left in the queue -- don't bother
+				# starting any more workers.
+				break
+
+			t = threading.Thread(target = self.worker, args = (i,))
+			t.start()
+			workers.append(t)
+		for worker in workers:
+			worker.join()
+		self.config.log("Thread farm finished with ", len(self.results), " results")
+		return self.results
 
 class Rawdog(Persistable):
 	"""The aggregator itself."""
@@ -632,8 +842,18 @@ class Rawdog(Persistable):
 	def __init__(self):
 		self.feeds = {}
 		self.articles = {}
+		self.plugin_storage = {}
 		self.state_version = STATE_VERSION
 
+	def get_plugin_storage(self, plugin):
+		try:
+			st = self.plugin_storage.setdefault(plugin, {})
+		except AttributeError:
+			# rawdog before 2.5 didn't have plugin storage.
+			st = {}
+			self.plugin_storage = {plugin: st}
+		return st
+
 	def check_state_version(self):
 		"""Check the version of the state file."""
 		try:
@@ -665,33 +885,56 @@ class Rawdog(Persistable):
 
 		print >>sys.stderr, "Feed URL automatically changed."
 
-	def list(self):
-		for url in self.feeds.keys():
-			feed = self.feeds[url]
+	def list(self, config):
+		"""List the configured feeds."""
+		for url, feed in self.feeds.items():
 			feed_info = feed.feed_info
 			print url
+			print "  ID:", feed.get_id(config)
 			print "  Hash:", short_hash(url)
-			print "  Title:", feed_info.get("title")
+			print "  Title:", feed.get_html_name(config)
 			print "  Link:", feed_info.get("link")
 
-	def update(self, config, feedurl = None):
-		config.log("Starting update")
-		now = time.time()
-
-		set_socket_timeout(config["timeout"])
-
+	def sync_from_config(self, config):
+		"""Update rawdog's internal state to match the
+		configuration."""
 		seenfeeds = {}
 		for (url, period, args) in config["feedslist"]:
 			seenfeeds[url] = 1
 			if not self.feeds.has_key(url):
 				config.log("Adding new feed: ", url)
 				self.feeds[url] = Feed(url)
-			self.feeds[url].period = period
-			self.feeds[url].args = args
+				self.modified()
+			feed = self.feeds[url]
+			if feed.period != period:
+				config.log("Changed feed period: ", url)
+				feed.period = period
+				self.modified()
+			newargs = {}
+			newargs.update(config["feeddefaults"])
+			newargs.update(args)
+			if feed.args != newargs:
+				config.log("Changed feed options: ", url)
+				feed.args = newargs
+				self.modified()
 		for url in self.feeds.keys():
 			if not seenfeeds.has_key(url):
 				config.log("Removing feed: ", url)
 				del self.feeds[url]
+				for key, article in self.articles.items():
+					if article.feed == url:
+						del self.articles[key]
+				self.modified()
+
+	def update(self, config, feedurl = None):
+		"""Perform the update action: check feeds for new articles, and
+		expire old ones."""
+		config.log("Starting update")
+		now = time.time()
+
+		feedparser._FeedParserMixin.can_contain_relative_uris = ["url"]
+		feedparser._FeedParserMixin.can_contain_dangerous_markup = []
+		set_socket_timeout(config["timeout"])
 
 		if feedurl is None:
 			update_feeds = [url for url in self.feeds.keys()
@@ -707,28 +950,43 @@ class Rawdog(Persistable):
 		numfeeds = len(update_feeds)
 		config.log("Will update ", numfeeds, " feeds")
 
+		if config["numthreads"] > 0:
+			fetcher = FeedFetcher(self, update_feeds, config)
+			prefetched = fetcher.run(config["numthreads"])
+		else:
+			prefetched = {}
+
 		count = 0
 		seen_some_items = {}
 		for url in update_feeds:
 			count += 1
 			config.log("Updating feed ", count, " of " , numfeeds, ": ", url)
-			if self.feeds[url].update(self, now, config):
+			feed = self.feeds[url]
+			if url in prefetched:
+				content = prefetched[url]
+			else:
+				plugins.call_hook("pre_update_feed", self, config, feed)
+				content = feed.fetch(self, config)
+			plugins.call_hook("mid_update_feed", self, config, feed, content)
+			rc = feed.update(self, now, config, content)
+			plugins.call_hook("post_update_feed", self, config, feed, rc)
+			if rc:
 				seen_some_items[url] = 1
 
 		count = 0
-		for key in self.articles.keys():
-			article = self.articles[key]
-			if ((not self.feeds.has_key(article.feed))
-			    or (seen_some_items.has_key(article.feed)
-			        and article.can_expire(now, config))):
+		for key, article in self.articles.items():
+			if (seen_some_items.has_key(article.feed)
+			    and article.can_expire(now, config)):
+				plugins.call_hook("article_expired", self, config, article, now)
 				count += 1
 				del self.articles[key]
-		config.log("Expired ", count, " articles, leaving ", len(self.articles.keys()))
+		config.log("Expired ", count, " articles, leaving ", len(self.articles))
 
 		self.modified()
 		config.log("Finished update")
 
 	def get_template(self, config):
+		"""Get the main template."""
 		if config["template"] != "default":
 			return load_file(config["template"])
 
@@ -737,6 +995,7 @@ class Rawdog(Persistable):
 <html lang="en">
 <head>
     <meta http-equiv="Content-Type" content="text/html; charset=ISO-8859-1">
+    <meta name="robots" content="noindex,nofollow,noarchive">
 """
 		if config["userefresh"]:
 			template += """__refresh__
@@ -762,17 +1021,19 @@ __feeds__
 <p id="aboutrawdog">Generated by
 <a href="http://offog.org/code/rawdog.html">rawdog</a>
 version __version__
-by <a href="mailto:azz@us-lot.org">Adam Sampson</a>.</p>
+by <a href="mailto:ats@offog.org">Adam Sampson</a>.</p>
 </div>
 </body>
-</html>"""
+</html>
+"""
 		return template
 
 	def get_itemtemplate(self, config):
+		"""Get the item template."""
 		if config["itemtemplate"] != "default":
 			return load_file(config["itemtemplate"])
 
-		template = """<div class="item feed-__feed_hash__" id="item-__hash__">
+		template = """<div class="item feed-__feed_hash__ feed-__feed_id__" id="item-__hash__">
 <p class="itemheader">
 <span class="itemtitle">__title__</span>
 <span class="itemfrom">[__feed_title__]</span>
@@ -786,66 +1047,111 @@ __description__
 		return template
 
 	def show_template(self, config):
+		"""Show the configured main template."""
 		print self.get_template(config)
 
 	def show_itemtemplate(self, config):
+		"""Show the configured item template."""
 		print self.get_itemtemplate(config)
 
-	def write(self, config):
-		outputfile = config["outputfile"]
-		config.log("Starting write")
-		now = time.time()
+	def write_article(self, f, article, config):
+		"""Write an article to the given file."""
+		feed = self.feeds[article.feed]
+		feed_info = feed.feed_info
+		entry_info = article.entry_info
 
-		bits = { "version" : VERSION }
+		link = entry_info.get("link")
+		if link == "":
+			link = None
 
-		refresh = config["expireage"]
-		for feed in self.feeds.values():
-			if feed.period < refresh: refresh = feed.period
+		guid = entry_info.get("id")
+		if guid == "":
+			guid = None
 
-		bits["refresh"] = """<meta http-equiv="Refresh" """ + 'content="' + str(refresh) + '"' + """>"""
+		itembits = {}
+		for name, value in feed.args.items():
+			if name.startswith("define_"):
+				itembits[name[7:]] = value
 
-		article_dates = {}
-		articles = self.articles.values()
-		for a in articles:
-			if config["sortbyfeeddate"]:
-				article_dates[a] = a.date or a.added
+		title = detail_to_html(entry_info.get("title_detail"), True, config)
+
+		key = None
+		for k in ["content", "summary_detail"]:
+			if entry_info.has_key(k):
+				key = k
+				break
+		if key is None:
+			description = None
+		else:
+			force_preformatted = feed.args.has_key("format") and (feed.args["format"] == "text")
+			description = detail_to_html(entry_info[key], False, config, force_preformatted)
+
+		date = article.date
+		if title is None:
+			if link is None:
+				title = "Article"
 			else:
-				article_dates[a] = a.added
-		numarticles = len(articles)
-		def compare(a, b):
-			"""Compare two articles to decide how they
-			   should be sorted. Sort by added date, then
-			   by feed, then by sequence, then by hash."""
-			i = cmp(article_dates[b], article_dates[a])
-			if i != 0:
-				return i
-			i = cmp(a.feed, b.feed)
-			if i != 0:
-				return i
-			i = cmp(a.sequence, b.sequence)
-			if i != 0:
-				return i
-			return cmp(a.hash, b.hash)
-		articles.sort(compare)
-		if config["maxarticles"] != 0:
-			articles = articles[:config["maxarticles"]]
+				title = "Link"
 
-		f = StringIO()
+		itembits["title_no_link"] = title
+		if link is not None:
+			itembits["url"] = string_to_html(link, config)
+		else:
+			itembits["url"] = ""
+		if guid is not None:
+			itembits["guid"] = string_to_html(guid, config)
+		else:
+			itembits["guid"] = ""
+		if link is None:
+			itembits["title"] = title
+		else:
+			itembits["title"] = '<a href="' + string_to_html(link, config) + '">' + title + '</a>'
+
+		itembits["feed_title_no_link"] = detail_to_html(feed_info.get("title_detail"), True, config)
+		itembits["feed_title"] = feed.get_html_link(config)
+		itembits["feed_url"] = string_to_html(feed.url, config)
+		itembits["feed_hash"] = short_hash(feed.url)
+		itembits["feed_id"] = feed.get_id(config)
+		itembits["hash"] = short_hash(article.hash)
+
+		if description is not None:
+			itembits["description"] = description
+		else:
+			itembits["description"] = ""
+
+		author = author_to_html(entry_info, feed.url, config)
+		if author is not None:
+			itembits["author"] = author
+		else:
+			itembits["author"] = ""
+
+		itembits["added"] = format_time(article.added, config)
+		if date is not None:
+			itembits["date"] = format_time(date, config)
+		else:
+			itembits["date"] = ""
+
+		plugins.call_hook("output_item_bits", self, config, feed, article, itembits)
 		itemtemplate = self.get_itemtemplate(config)
-		dw = DayWriter(f, config)
+		f.write(fill_template(itemtemplate, itembits))
 
+	def write_remove_dups(self, articles, config, now):
+		"""Filter the list of articles to remove articles that are too
+		old or are duplicates."""
+		kept_articles = []
 		seen_links = {}
 		seen_guids = {}
-
-		count = 0
 		dup_count = 0
 		for article in articles:
-			age = now - article.added
-			if config["maxage"] != 0 and age > config["maxage"]:
-				break
-
 			feed = self.feeds[article.feed]
-			feed_info = feed.feed_info
+			age = now - article.added
+
+			maxage = config["maxage"]
+			if "maxage" in feed.args:
+				maxage = feed.args["maxage"]
+			if maxage != 0 and age > maxage:
+				continue
+
 			entry_info = article.entry_info
 
 			link = entry_info.get("link")
@@ -873,64 +1179,20 @@ __description__
 					dup_count += 1
 					continue
 
-			count += 1
-			dw.time(article_dates[article])
+			kept_articles.append(article)
+		return (kept_articles, dup_count)
 
-			itembits = {}
+	def get_main_template_bits(self, config):
+		"""Get the bits that are used in the default main template,
+		with the exception of items and num_items."""
+		bits = { "version" : VERSION }
+		bits.update(config["defines"])
 
-			title = detail_to_html(entry_info.get("title_detail"), True, config)
+		refresh = config["expireage"]
+		for feed in self.feeds.values():
+			if feed.period < refresh: refresh = feed.period
 
-			key = None
-			for k in ["content", "summary_detail"]:
-				if entry_info.has_key(k):
-					key = k
-					break
-			if key is None:
-				description = None
-			else:
-				force_preformatted = feed.args.has_key("format") and (feed.args["format"] == "text")
-				description = detail_to_html(entry_info[key], False, config, force_preformatted)
-
-			date = article.date
-			if title is None:
-				if link is None:
-					title = "Article"
-				else:
-					title = "Link"
-
-			itembits["title_no_link"] = title
-			if link is not None:
-				itembits["url"] = link
-			else:
-				itembits["url"] = ""
-			if link is None:
-				itembits["title"] = title
-			else:
-				itembits["title"] = '<a href="' + url_to_html(link) + '">' + title + '</a>'
-
-			itembits["feed_title_no_link"] = detail_to_html(feed_info.get("title_detail"), True, config)
-			itembits["feed_title"] = feed.get_html_link(config)
-			itembits["feed_url"] = feed.url
-			itembits["feed_hash"] = short_hash(feed.url)
-			itembits["hash"] = short_hash(article.hash)
-
-			if description is not None:
-				itembits["description"] = description
-			else:
-				itembits["description"] = ""
-
-			itembits["added"] = format_time(article.added, config)
-			if date is not None:
-				itembits["date"] = format_time(date, config)
-			else:
-				itembits["date"] = ""
-
-			f.write(fill_template(itemtemplate, itembits))
-
-		dw.close()
-		bits["items"] = f.getvalue()
-		bits["num_items"] = str(numarticles)
-		config.log("Selected ", count, " of ", numarticles, " articles to write; ignored ", dup_count, " duplicates")
+		bits["refresh"] = """<meta http-equiv="Refresh" """ + 'content="' + str(refresh) + '"' + """>"""
 
 		f = StringIO()
 		print >>f, """<table id="feeds">
@@ -950,16 +1212,86 @@ __description__
 		bits["feeds"] = f.getvalue()
 		bits["num_feeds"] = str(len(feeds))
 
+		return bits
+
+	def write_output_file(self, articles, article_dates, config):
+		"""Write a regular rawdog HTML output file."""
+		f = StringIO()
+		dw = DayWriter(f, config)
+		plugins.call_hook("output_items_begin", self, config, f)
+
+		for article in articles:
+			if not plugins.call_hook("output_items_heading", self, config, f, article, article_dates[article]):
+				dw.time(article_dates[article])
+
+			self.write_article(f, article, config)
+
+		dw.close()
+		plugins.call_hook("output_items_end", self, config, f)
+
+		bits = self.get_main_template_bits(config)
+		bits["items"] = f.getvalue()
+		bits["num_items"] = str(len(self.articles))
+		plugins.call_hook("output_bits", self, config, bits)
 		s = fill_template(self.get_template(config), bits)
+		outputfile = config["outputfile"]
 		if outputfile == "-":
-			print s
+			write_ascii(sys.stdout, s, config)
 		else:
 			config.log("Writing output file: ", outputfile)
 			f = open(outputfile + ".new", "w")
-			print >>f, s
+			write_ascii(f, s, config)
 			f.close()
 			os.rename(outputfile + ".new", outputfile)
 
+	def write(self, config):
+		"""Perform the write action: write articles to the output
+		file."""
+		config.log("Starting write")
+		now = time.time()
+
+		article_dates = {}
+		articles = self.articles.values()
+		for a in articles:
+			if config["sortbyfeeddate"]:
+				article_dates[a] = a.date or a.added
+			else:
+				article_dates[a] = a.added
+		numarticles = len(articles)
+
+		def compare(a, b):
+			"""Compare two articles to decide how they
+			   should be sorted. Sort by added date, then
+			   by feed, then by sequence, then by hash."""
+			i = cmp(article_dates[b], article_dates[a])
+			if i != 0:
+				return i
+			i = cmp(a.feed, b.feed)
+			if i != 0:
+				return i
+			i = cmp(a.sequence, b.sequence)
+			if i != 0:
+				return i
+			return cmp(a.hash, b.hash)
+		plugins.call_hook("output_filter", self, config, articles)
+		articles.sort(compare)
+		plugins.call_hook("output_sort", self, config, articles)
+
+		if config["maxarticles"] != 0:
+			articles = articles[:config["maxarticles"]]
+
+		plugins.call_hook("output_write", self, config, articles)
+
+		if not plugins.call_hook("output_sorted_filter", self, config, articles):
+			(articles, dup_count) = self.write_remove_dups(articles, config, now)
+		else:
+			dup_count = 0
+
+		config.log("Selected ", len(articles), " of ", numarticles, " articles to write; ignored ", dup_count, " duplicates")
+
+		if not plugins.call_hook("output_write_files", self, config, articles, article_dates):
+			self.write_output_file(articles, article_dates, config)
+
 		config.log("Finished write")
 
 def usage():
@@ -970,6 +1302,7 @@ Usage: rawdog [OPTION]...
 General options (use only once):
 -d|--dir DIR                 Use DIR instead of ~/.rawdog
 -v, --verbose                Print more detailed status information
+-N, --no-locking             Do not lock the state file
 --help                       Display this help and exit
 
 Actions (performed in order given):
@@ -987,13 +1320,13 @@ Special actions (all other options are ignored if one of these is specified):
 --upgrade OLDDIR NEWDIR      Import feed state from rawdog 1.x directory
                              OLDDIR into rawdog 2.x directory NEWDIR
 
-Report bugs to <azz@us-lot.org>."""
+Report bugs to <ats@offog.org>."""
 
 def main(argv):
 	"""The command-line interface to the aggregator."""
 
 	try:
-		(optlist, args) = getopt.getopt(argv, "ulwf:c:tTd:va:", ["update", "list", "write", "update-feed=", "help", "config=", "show-template", "dir=", "show-itemtemplate", "verbose", "upgrade", "add="])
+		(optlist, args) = getopt.getopt(argv, "ulwf:c:tTd:va:N", ["update", "list", "write", "update-feed=", "help", "config=", "show-template", "dir=", "show-itemtemplate", "verbose", "upgrade", "add=", "no-locking"])
 	except getopt.GetoptError, s:
 		print s
 		usage()
@@ -1010,6 +1343,7 @@ def main(argv):
 
 	statedir = os.environ["HOME"] + "/.rawdog"
 	verbose = 0
+	locking = 1
 	for o, a in optlist:
 		if o == "--help":
 			usage()
@@ -1018,6 +1352,8 @@ def main(argv):
 			statedir = a
 		elif o in ("-v", "--verbose"):
 			verbose = 1
+		elif o in ("-N", "--no-locking"):
+			locking = 0
 
 	try:
 		os.chdir(statedir)
@@ -1032,11 +1368,14 @@ def main(argv):
 		print >>sys.stderr, "In config:"
 		print >>sys.stderr, err
 		return 1
-	config["verbose"] = verbose
+	if verbose:
+		config["verbose"] = True
 
-	persister = Persister("state", Rawdog)
+	persister = Persister("state", Rawdog, locking)
 	try:
 		rawdog = persister.load()
+	except KeyboardInterrupt:
+		return 1
 	except:
 		print "An error occurred while reading state from " + statedir + "/state."
 		print "This usually means the file is corrupt, and removing it will fix the problem."
@@ -1048,15 +1387,19 @@ def main(argv):
 		print "Removing the state file will fix it."
 		return 1
 
+	rawdog.sync_from_config(config)
+
+	plugins.call_hook("startup", rawdog, config)
+
 	for o, a in optlist:
 		if o in ("-u", "--update"):
 			rawdog.update(config)
-		elif o in ("-l", "--list"):
-			rawdog.list()
-		elif o in ("-w", "--write"):
-			rawdog.write(config)
 		elif o in ("-f", "--update-feed"):
 			rawdog.update(config, a)
+		elif o in ("-l", "--list"):
+			rawdog.list(config)
+		elif o in ("-w", "--write"):
+			rawdog.write(config)
 		elif o in ("-c", "--config"):
 			try:
 				config.load(a)
@@ -1070,6 +1413,10 @@ def main(argv):
 			rawdog.show_itemtemplate(config)
 		elif o in ("-a", "--add"):
 			add_feed("config", a, config)
+			config.reload()
+			rawdog.sync_from_config(config)
+
+	plugins.call_hook("shutdown", rawdog, config)
 
 	persister.save()
 
